#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include "keyboard.h"

#define BAT_OK    (1<<2) // not used

// *** special keycodes (8,9,10,27, 140-179 are not supposed to be printed) ***
#define KEY_BACK  8
#define KEY_TAB   9
#define KEY_ENTER 10
#define KEY_ESC   27
#define KEY_ENTF  140
#define KEY_CLEAR 141
#define KEY_ENDE  142
#define KEY_POS1  143
#define KEY_EINFG 144
#define KEY_UP    150
#define KEY_DOWN  151
#define KEY_LEFT  152
#define KEY_RIGHT 153
#define KEY_PGUP  154
#define KEY_PGDN  155
#define KEY_F1    156
#define KEY_F2    157
#define KEY_F3    158
#define KEY_F4    159
#define KEY_F5    160
#define KEY_F6    161
#define KEY_F7    162
#define KEY_F8    163
#define KEY_F9    164
#define KEY_F10   165
#define KEY_F11   166
#define KEY_F12   167
#define KEY_AE    196 
#define KEY_OE    214 
#define KEY_UE    220


// Unshifted characters
const unsigned char unshifted[][2] PROGMEM =
{
	{0x05,KEY_F1},
	{0x06,KEY_F2},
	{0x04,KEY_F3},
	{0x0c,KEY_F4},
	{0x03,KEY_F5},
	{0x0b,KEY_F6},
	{0x83,KEY_F7},
	{0x0a,KEY_F8},
	{0x01,KEY_F9},
	{0x09,KEY_F10},
	{0x78,KEY_F11},
	{0x07,KEY_F12},
	{0x0d,KEY_TAB},
	{0x0e,'^'},
	{0x15,'q'},
	{0x16,'1'},
	{0x1a,'y'},
	{0x1b,'s'},
	{0x1c,'a'},
	{0x1d,'w'},
	{0x1e,'2'},
	{0x21,'c'},
	{0x22,'x'},
	{0x23,'d'},
	{0x24,'e'},
	{0x25,'4'},
	{0x26,'3'},
	{0x29,' '},
	{0x2a,'v'},
	{0x2b,'f'},
	{0x2c,'t'},
	{0x2d,'r'},
	{0x2e,'5'},
	{0x31,'n'},
	{0x32,'b'},
	{0x33,'h'},
	{0x34,'g'},
	{0x35,'z'},
	{0x36,'6'},
	{0x39,','},
	{0x3a,'m'},
	{0x3b,'j'},
	{0x3c,'u'},
	{0x3d,'7'},
	{0x3e,'8'},
	{0x41,','},
	{0x42,'k'},
	{0x43,'i'},
	{0x44,'o'},
	{0x45,'0'},
	{0x46,'9'},
	{0x49,'.'},
	{0x4a,'-'},
	{0x4b,'l'},
	{0x4c,''},
	{0x4d,'p'},
	{0x4e,''},
	{0x52,''},
	{0x54,''},
	{0x55,'@'},
	{0x5a,KEY_ENTER},
	{0x5b,'+'},
	{0x5d,'#'},
	{0x61,'<'},
	{0x66,KEY_BACK},
	{0x69,'1'},
	{0x6b,'4'},
	{0x6c,'7'},
	{0x70,'0'},
	{0x71,','},
	{0x72,'2'},
	{0x73,'5'},
	{0x74,'6'},
	{0x75,'8'},
	{0x76,KEY_ESC},
	{0x79,'+'},
	{0x7a,'3'},
	{0x7b,'-'},
	{0x7c,'*'},
	{0x7d,'9'},
	{0,0}
};

// Shifted characters
const unsigned char shifted[][2] PROGMEM =
{
	{0x0d,9},
	{0x0e,''},
	{0x15,'Q'},
	{0x16,'!'},
	{0x1a,'Y'},
	{0x1b,'S'},
	{0x1c,'A'},
	{0x1d,'W'},
	{0x1e,'"'},
	{0x21,'C'},
	{0x22,'X'},
	{0x23,'D'},
	{0x24,'E'},
	{0x25,'$'},
	{0x26,''},
	{0x29,' '},
	{0x2a,'V'},
	{0x2b,'F'},
	{0x2c,'T'},
	{0x2d,'R'},
	{0x2e,'%'},
	{0x31,'N'},
	{0x32,'B'},
	{0x33,'H'},
	{0x34,'G'},
	{0x35,'Z'},
	{0x36,'&'},
	{0x39,'L'},
	{0x3a,'M'},
	{0x3b,'J'},
	{0x3c,'U'},
	{0x3d,'/'},
	{0x3e,'('},
	{0x41,';'},
	{0x42,'K'},
	{0x43,'I'},
	{0x44,'O'},
	{0x45,'='},
	{0x46,')'},
	{0x49,':'},
	{0x4a,'_'},
	{0x4b,'L'},
	{0x4c,''},
	{0x4d,'P'},
	{0x4e,'?'},
	{0x52,''},
	{0x54,''},
	{0x55,'`'},
	{0x5a,KEY_ENTER},
	{0x5b,'*'},
	{0x5d,''},
	{0x61,'>'},
	{0x66,8},
	{0x69,'1'},
	{0x6b,'4'},
	{0x6c,'7'},
	{0x70,'0'},
	{0x71,','},
	{0x72,'2'},
	{0x73,'5'},
	{0x74,'6'},
	{0x75,'8'},
	{0x79,'+'},
	{0x7a,'3'},
	{0x7b,'-'},
	{0x7c,'*'},
	{0x7d,'9'},
	{0,0}
};

// extended keys
const unsigned char extended[][2] PROGMEM =
{
	{0x6b,KEY_LEFT},  
	{0x74,KEY_RIGHT},  
	{0x75,KEY_UP},  
	{0x72,KEY_DOWN},  
	{0x70,KEY_EINFG},  
	{0x71,KEY_ENTF},  
	{0x6c,KEY_POS1},  
	{0x69,KEY_ENDE},  
	{0x7d,KEY_PGUP},
	{0x7a,KEY_PGDN},
	{0,0}
};

unsigned char kb_buffer[KB_BUFF_SIZE];	// keyboard buffer
unsigned char *input, *output;			// pointers to kb_buffer[]
unsigned char buffcnt;					// number of bytes stored in kb_buffer[]
unsigned char sendbyte = 0;				// byte to be sent to the keyboard
unsigned char mode = 0;					// scan code mode flag (1 means we send the raw scan codes to USART instead of decoding them, for debugging purposes)       
unsigned char parity;					// parity bit
unsigned char bg;						// stores ACK bit of host -> keyboard transfer
unsigned char edge, bitcount;			// edge: bit 0: 0 = falling edge, 1 = rising edge
										//       bit 1: 0 = receiving from, 1 = sending to keyboard
unsigned char KB_BAT;					// set to 1 when "BAT successful" code 0xAA was received
unsigned char div_flags;
volatile unsigned char has_aborted;

void put_kbbuff(unsigned char c);
void decode(unsigned char sc);

void KB_init(void)
{
	// Initialise buffers
	input    = kb_buffer;
	output   = kb_buffer;
	buffcnt = 0;
	// INT1 interrupt on falling edge
	MCUCR  |= (2<<ISC10);
	GICR   |= (1<<INT1);
	mode    = 0; 
	edge    = 0;          
	bitcount= 11;
	parity  = 0;
	
	sei();
}

void KB_send(unsigned char data)
{
	// send [data] to keyboard
	unsigned char i;
	if(bitcount != 11) return;			// abort if already transmitting / receiving
	GICR		&= ~(1<<INT1);			// disable INT1
	KEY_PORT	&= ~_BV(KEY_CLOCK);	// CLOCK low
	KEY_PORT	|=  _BV(KEY_DATA);		// DATA high
	KEY_DDR		|=  _BV(KEY_CLOCK);		// Port D->KEY_CLOCK output
	KEY_DDR		|=  _BV(KEY_DATA);		// Port D->KEY_DATA output

	// wait 60 s = 960 cycles
	for (i=0;i<240;i++);

	KEY_PORT	&= ~_BV(KEY_DATA);		// DATA low
	KEY_PORT	|=  _BV(KEY_CLOCK);		// CLOCK high
	KEY_DDR		&= ~_BV(KEY_CLOCK);	// Port D->KEY_CLOCK input     
	GIFR		|= (1<<INTF1);			// clear GIFR->INT1F by setting it to 1
	
	sendbyte = data;
	bitcount = 11;
	parity   = 0;
	edge     = 2;					// 0 = falling edge 
	MCUCR   |=  (2<<ISC10);		// INT1 interrupt on falling edge
	GICR    |=  (1<<INT1);			// enable INT1
}

unsigned char KB_getchar(void)
{
	// read out the next key press from kb_buffer[]
	int byte;
	if(buffcnt == 0) return 0;
	byte = *output;// Get byte
	output++; // Increment pointer
	if(output >= kb_buffer + KB_BUFF_SIZE) // Pointer wrapping
		output = kb_buffer;
	buffcnt--; // Decrement buffer count
	return byte;
}

ISR(INT1_vect)
{
	static unsigned char data;	// Holds the received scan code
	if(!(edge & 1))
	{
		// Routine entered at falling edge
		if(!(edge & 2))
		{
			// * receive *
			if(bitcount < 11 && bitcount > 2)
			{
				// Bit 3 to 10 is data. Start and stop bits are ignored.
				data = (data >> 1);
				if(KEY_PIN & _BV(KEY_DATA))
				{
					// Store a '1'
					data = data | 0x80;
					parity ^= 1; 
				}
			}
			// odd parity
			if(bitcount == 2)
				// thus: parity ok <=> parity bit is 0 
				if(!(KEY_PIN & _BV(KEY_DATA)))
					parity ^= 1;
		}
		else
		{
			// *  send   * 
			if(bitcount < 12 && bitcount > 3)
			{
				if(sendbyte & 1)
				{
					// send a '1'
					KEY_PORT |= _BV(KEY_DATA);	
					parity++;
				}
				else
					// send a '0'
					KEY_PORT &= ~_BV(KEY_DATA);
				sendbyte >>= 1;
			}
			else if(bitcount == 3)
			{
				// generate odd parity
				if(parity & 1)  
					KEY_PORT &= ~_BV(KEY_DATA);	
				else
					KEY_PORT |= _BV(KEY_DATA);	  
				parity = 0;
			}
			else if(bitcount == 2)
			{
				// release DATA pin
				KEY_PORT |=  _BV(KEY_DATA);
				KEY_DDR  &= ~_BV(KEY_DATA);  
			}
			else if(bitcount == 1)
			{
				// read out ACK bit
				if(!(KEY_PIN & _BV(KEY_DATA)))
					// failed
					bg = 0;
			}
		}
		// Set interrupt on rising edge
		MCUCR |= (1<<ISC10);	
		edge ^= 1;
	}
	else
	{
		// Routine entered at rising edge
		bitcount--;
		// Set interrupt on falling edge
		MCUCR &= ~(1<<ISC10);
		edge  ^= 1;
		if(bitcount == 0)
		{
			// All bits received
			bitcount = 11;
			if(!(edge&2))
			{
				#ifdef USE_KB_PARITY
				if(parity)
				{ 
					//printUSART("kb:p err");
					KB_init();
				}
				else
				#endif
					decode(data);
			}
			// clear send flag
			else edge &= 1;
		}
	}
}

void decode(unsigned char sc)
{
	// decode scancodes 
	static unsigned char is_up=0, shift = 0, ext = 0, caps = 0; // flags
	static unsigned char sendnext = 0; // byte to be sent next
	unsigned char i;
	if(!is_up)
	{
		// Last data received was the up-key identifier
		switch (sc)
		{
			// The up-key identifier
			case 0xF0:
				is_up = 1;
				ext   = 0;
				break;
			case 0x12: // Left SHIFT
			case 0x59: // Right SHIFT
				shift = 1;
				break;
			case 0x58: // Caps lock
				caps = ~caps;
				// switch LED on / off
				KB_send(0xed); 
				sendnext = 0x04 & caps;
				break;
			case 0xe0: // extended code
				ext = 1;
				break;
			case 0x05: // F1: enter/leave scan code mode
				#ifdef DEBUG
				if (mode == 0)
					mode = 1;// Enter 
				if (mode == 2)
					mode = 3;// Leave
				#endif 
				break;
			case 0x07: // F12: turn LCD on / off
				/*if(div_flags & LCD_AN)
				{
					LCDoff();
					div_flags &= ~LCD_AN;
				}
				else
				{
					LCDon();
					div_flags |= LCD_AN;
				}*/
				break;
			case 0xAA: // BAT successful
				div_flags |= BAT_OK;
				break;
			case 0xFA: // got ACK -> next byte to send? 
				if(sendnext!=0xf1)
				{
					KB_send(sendnext);
					// 0xF1 = invalid command; tells us we do not need to send another byte when we get ACK
					sendnext = 0xf1;
				}
				break; 
			#ifdef DEBUG
			case 0xee: // command 0xEE is simply echoed by the keyboard 
				//printUSART("KB_ECHO");
				break;
			#endif
			case 0x76: // ESC: set abort flag
				has_aborted = 1;
				// no break!
			default:
				if(mode == 0 || mode == 3)
				{
					// If ASCII mode
					if(ext)
					{
						// extended scan code
						for(i = 0; pgm_read_byte(&extended[i][0]) != sc && pgm_read_byte(&extended[i][0]); i++);
						if(pgm_read_byte(&extended[i][0]) == sc)
							put_kbbuff(pgm_read_byte(&extended[i][1]));
					}
					else if(!(shift | caps))
					{
						// If shift not pressed,
						// do a table look-up
						for(i = 0; pgm_read_byte(&unshifted[i][0]) != sc && pgm_read_byte(&unshifted[i][0]); i++);
						if(pgm_read_byte(&unshifted[i][0]) == sc)
							put_kbbuff(pgm_read_byte(&unshifted[i][1]));
					}
					else
					{
						// If shift pressed
						for(i = 0; pgm_read_byte(&shifted[i][0]) != sc && pgm_read_byte(&shifted[i][0]); i++);
						if(pgm_read_byte(&shifted[i][0]) == sc)
							put_kbbuff(pgm_read_byte(&shifted[i][1]));
					}
			}
			else
			{
				// Scan code mode
				UDR = sc; // send scan code to USART
			}
			break;
		}
	}
	else
	{
		is_up = 0;// Two 0xF0 in a row not allowed
		switch(sc)
		{
			case 0x12 :// Left SHIFT
				shift = 0;
				break;
			case 0x59 :// Right SHIFT
				shift = 0;
				break;
			case 0x05 :// F1
				#ifdef DEBUG
				if(mode == 1)
					mode = 2;
				if(mode == 3)
					mode = 0;
				#endif
				break;
		}
	}
}

void put_kbbuff(unsigned char c)
{
	// puts a decoded key code into kb_buffer[]
	if (buffcnt<KB_BUFF_SIZE)
	{
		// If buffer not full
		*input = c;// Put character into buffer
		input++; // Increment pointer
		buffcnt++;
		if(input >= kb_buffer + KB_BUFF_SIZE)// Pointer wrapping
		input = kb_buffer;
	}
}

