//**************************************************
// ***** SOURCE FILE : FAT.c ******
//**************************************************

#include <avr/io.h>
#include <stdio.h>
#include <string.h>
#include "FAT.h"
#include "SD.h"

#define FIRST_SECTOR(cluster)	\
	(((cluster - 2) * sectorPerCluster) + firstDataSector)

unsigned long startBlock;
unsigned long firstDataSector;
unsigned long rootCluster;
unsigned long totalClusters;
unsigned long bootSector;
unsigned long currentCluster;
unsigned int bytesPerSector;
unsigned int sectorPerCluster;
unsigned int reservedSectorCount;


unsigned long lastUsedCluster(void);
void convertFileName(const char *src, char *dst);


unsigned char FS_init(void)
{
	struct bootsector_t *bpb; //mapping the buffer onto the structure
	struct mbrinfo_t *mbr;
	struct partitioninfo_t *partition;
	struct fsinfo_t *fs = (struct fsinfo_t *)&buffer;
	unsigned long dataSectors;

	bootSector = 0;

	SD_read(0, buffer);
	bpb = (struct bootsector_t *)buffer;

	if(bpb->jumpBoot[0]!=0xE9 && bpb->jumpBoot[0]!=0xEB) //check if it is boot sector
	{
		mbr = (struct mbrinfo_t *) buffer;         //if it is not boot sector, it must be MBR

		if(mbr->signature != 0xaa55) return 0;            //if it is not even MBR then it's not FAT32

		partition = (struct partitioninfo_t *)(mbr->partitionData); //first partition
		bootSector = partition->firstSector; //the unused sectors, hidden to the FAT

		SD_read(partition->firstSector, buffer);//read the bpb sector
		bpb = (struct bootsector_t *)buffer;
		if(bpb->jumpBoot[0]!=0xE9 && bpb->jumpBoot[0]!=0xEB)  return 0; 
	}
	
	bytesPerSector      = bpb->bytesPerSector;
	sectorPerCluster    = bpb->sectorPerCluster;
	reservedSectorCount = bpb->reservedSectorCount;
	rootCluster         = bpb->rootCluster;// + (sector / sectorPerCluster) +1;
	firstDataSector     = bpb->hiddenSectors + bootSector + reservedSectorCount + (bpb->numberofFATs * bpb->FATsize_F32);

	dataSectors   = bpb->totalSectors_F32 - bpb->reservedSectorCount - ( bpb->numberofFATs * bpb->FATsize_F32);
	totalClusters = dataSectors / sectorPerCluster;
	
	currentCluster = rootCluster;
	
	SD_read(bootSector + 1, buffer);
	if((fs->leadSignature != 0x41615252) || (fs->structureSignature != 0x61417272) || (fs->trailSignature !=0xaa550000))
		return 0;
		
	#ifdef FAT_DEBUG
	printf("Bytes/Sector:		%d\n", bytesPerSector);
	printf("Sectors/cluster:	%d\n", sectorPerCluster);
	printf("Reserved sectors:	%d\n", reservedSectorCount);
	printf("Root cluster:		%lu\n", rootCluster);
	printf("First Sector:		%lu\n", firstDataSector);
	printf("Sectors:			%lu\n", dataSectors);
	printf("Clusters:			%lu\n", totalClusters);
	printf("Boot sector:		%lu\n", bootSector);
	#endif
		
	return 1;
}

unsigned long FS_fatentry(unsigned long cluster, unsigned long *newval)
{
	unsigned long sector = bootSector + reservedSectorCount + (cluster * 4) / bytesPerSector;
	unsigned int offset = (cluster * 4) % bytesPerSector;
	
	SD_read(sector, buffer);
	
	if(newval)
	{
		unsigned long oldval = *(unsigned long *)&buffer[offset];
		*newval &= 0x0FFFFFFF;
		*(unsigned long *)&buffer[offset] &= 0xF0000000;
		*(unsigned long *)&buffer[offset] |= *newval;
		SD_write(sector, buffer);
		return oldval;
	}
	else return (*(unsigned long *)&buffer[offset] & 0x0FFFFFFF);
}

unsigned long FS_firstSector(unsigned long cluster)
{
	return FIRST_SECTOR(cluster);
}

unsigned long FS_nextFreeCluster(unsigned long startCluster)
{
	unsigned long cluster, *value, sector;
	unsigned char i;

	if(!startCluster) startCluster = lastUsedCluster() + 1;
	startCluster -=  (startCluster % 128);   //to start with the first file in a FAT sector
	for(cluster = startCluster; cluster < totalClusters; cluster += 128)
	{
		sector = bootSector + reservedSectorCount + ((cluster * 4) / bytesPerSector);
		SD_read(sector, buffer);
		for(i= 0; i< 128; i++)
		{
			value = (unsigned long *) &buffer[i*4];
			if(((*value) & 0x0fffffff) == 0)
			return cluster + i;
		}
	}

	return 0;
}

struct direntry_t *FS_direntry(const char *filename, unsigned long *sector)
{
	struct direntry_t *de = 0;
	struct ldirentry_t *we;
	unsigned char nameMatch,longName = 0;
	unsigned char i, index = 16;
	char *nameptr;
	
	unsigned long sect = FIRST_SECTOR(currentCluster);
	
	while(1)
	{
		if(index == 16)	// time for next sector ?
		{
			SD_read(sect++, buffer);
			de = (struct direntry_t *)buffer;
			index = 0;
		}
		
		nameMatch = 1;
		
		// check the status of this directory entry slot
		if(de->name[0] == 0x00)
		{
			// slot is empty and this is the end of directory
			return NULL;
		}
		else if(de->name[0] == 0xE5 && !filename)
		{
			// this is an empty slot
			if(sector) *sector = sect - 1;
			return de;
		}
		else if(filename)
		{
			// this is a valid and occupied entry
			// is it a part of a long file/dir name?
			if(de->attrib == ATTR_LONG_NAME)
			{
				// we have a long name entry
				// cast this directory entry as a "windows" (LFN: LongFileName) entry
				we = (struct ldirentry_t *)de;
				
				nameptr = (char *)&filename[13 * ((we->ord-1) & 0x0f)];
				for(i = 0; i < 5; i++)
				{
					if(we->name1[i*2] == 0xFF || we->name1[i*2] == '\n') break;
					if(*nameptr++ != we->name1[i*2]) { nameMatch = 0; break; }
				}
				for(i = 0; i < 6; i++)
				{
					if(we->name2[i*2] == 0xFF || we->name2[i*2] == '\n') break;
					if(*nameptr++ != we->name2[i*2]) { nameMatch = 0; break; }
				}
				for(i = 0; i < 2; i++)
				{
					if(we->name3[i*2] == 0xFF || we->name3[i*2] == '\n') break;
					if(*nameptr++ != we->name3[i*2]) { nameMatch = 0; break; }
				}
				
				if(!nameMatch)
				{
					de++; index++;
					continue;
				}
				if(we->ord & 0x40) longName = we->ord & 0x0f;
				if((we->ord & 0x0f) == longName - 1) longName--;
			}
			else
			{
				// we have a short name entry
				
				// check if this is the short name entry corresponding
				// to the end of a multi-part long name entry
				if(longName == 1)
				{
					// a long entry name has been collected
					if(sector) *sector = sect - 1;
					return de;
				}
				else
				{
					char tmp[11];
					convertFileName(filename, tmp);
					
					// entry is a short name (8.3 format) without a
					// corresponding multi-part long name entry
					nameptr = &tmp[0];
					for(i = 0; i < 8; i++)	if(*nameptr++ != de->name[i]) { nameMatch = 0; break; }
					for(i = 0; i < 3; i++)	if(*nameptr++ != de->name[i+8]) { nameMatch = 0; break; }
					if(!nameMatch)
					{
						de++; index++;
						continue;
					}
					if(sector) *sector = sect - 1;
					return de;
				}
			}
		}
		de++; index++;
	}
	
	return NULL;
}

unsigned char FS_changedir(const char *dirname)
{
	struct direntry_t *de = FS_direntry(dirname, NULL);
	if(!de)
	{
		#ifdef FAT_DEBUG
		printf("Cannot find the path specified.\n");
		#endif
		return 0;
	}
	
	currentCluster = (unsigned long)de->firstClusterHI << 16 | de->firstClusterLO;
	
	return 1;
}


unsigned long lastUsedCluster(void)
{
	struct fsinfo_t *fs = (struct fsinfo_t *)&buffer;
	SD_read(bootSector + 1, buffer);
	
	return fs->nextFreeCluster;
}

void convertFileName(const char *src, char *dst)
{
	int j, k;
	
	for(j = 0; j < 11; j++)
		if(src[j] == '.' || src[j] == 0) break;

	if(j > 8) return;

	for(k = 0; k < j; k++) //setting file name
		dst[k] = src[k];

	for(k = j; k <= 7; k++) //filling file name trail with blanks
		dst[k] = ' ';

	if(src[j] == '.') j++;
	for(k = 8; k < 11; k++) //setting file extention
	{
		if(src[j] != 0)
			dst[k] = src[j++];
		else while(k < 11)
			dst[k++] = ' ';
	}

	for(j = 0; j < 11; j++) //converting small letters to caps
		if((dst[j] >= 0x61) && (dst[j] <= 0x7a))
			dst[j] -= 0x20;
}
