#include <avr/io.h>
#include <stdio.h>
#include "SD.h"

#ifndef F_CPU
#define F_CPU	16000000LU
#endif
#include <util/delay.h>
	
#ifdef SD_DEBUG
#define PRINT(STR)					\
	printf("%s\n", STR)
#else
#define PRINT(STR)
#endif

#define CHECK(cond)					\
do {								\
	if(!(cond))						\
	{								\
		PRINT(#cond);				\
		SD_PORT |= _BV(SD_CS_PIN);	\
		return 0;					\
	}								\
} while(0)


static unsigned char SD_response;
unsigned char spisend(unsigned char data);

unsigned char SD_init(void)
{
	// initialize SPI interface
	SD_PORT = _BV(SD_CS_PIN) | _BV(SD_MISO_PIN);
	SD_DDR = _BV(SD_CS_PIN) | _BV(SD_MOSI_PIN) | _BV(SD_SCK_PIN);
	
	#ifdef SD_SPI_CONNECTION
	SPCR = _BV(SPE) | _BV(MSTR);
	SPSR |= _BV(SPI2X);
	#endif
	
	unsigned int i = 0;
	
	do
	{
		i++;
		_delay_ms(15);
		
		// resetting card, go to SPI mode
		SD_command(SD_GO_IDLE_STATE, 0);
		
	} while(SD_response != SD_R1_IDLE_STATE && i < 10);
	CHECK(i < 10);
	
	i = 0;
	do
	{
		i++;
		// initializing card for operation
		SD_command(SD_SEND_OP_COND, 0);
	} while(SD_response && i < 10);
	CHECK(i < 10);
	
	CHECK(SD_command(SD_CRC_ON_OFF, 0) != 0);
	CHECK(SD_command(SD_SET_BLOCKLEN, SD_BLOCKSIZE) != 0);
	
	return 1;
}


unsigned char SD_command(unsigned char cmd, unsigned long arg)
{
	unsigned int i = 0;

	// assert chip select
	SD_PORT &= ~_BV(SD_CS_PIN);
	
	spisend(cmd | 0x40);
	spisend((unsigned char)(arg >> 24));
	spisend((unsigned char)(arg >> 16));
	spisend((unsigned char)(arg >> 8));
	spisend((unsigned char)arg);
	spisend(0x95);	// crc valid only for SD_GO_IDLE_STATE
	// end command
	// wait for response
	// if more than 8 retries, card has timed-out
	
	do
	{
		i++;
		SD_response = spisend(0xFF);
	} while(SD_response == 0xFF && i < 8);
	CHECK(i < 8);
	
	_delay_ms(10);
	
	#ifdef SD_DEBUG
	printf("SD_response = 0x%x\n", SD_response);
	#endif
	
	spisend(0xFF); //extra 8 CLK
	
	SD_PORT |= _BV(SD_CS_PIN);

	return 1;
}

unsigned char SD_read(unsigned long sector, unsigned char *buffer)
{
	unsigned int i = 0;
	unsigned char tmp;

	// assert chip select
	// issue command
	CHECK(SD_command(SD_READ_SINGLE_BLOCK, sector << 9) != 0);
	
	SD_PORT &= ~_BV(SD_CS_PIN);
	// wait for block start
	do
	{
		i++;
		tmp = spisend(0xFF);
	} while(tmp != SD_STARTBLOCK_READ);
	
	// read in data
	for(i = 0; i < SD_BLOCKSIZE; i++)
		buffer[i] = spisend(0xFF);
	
	// read 16-bit CRC
	spisend(0xFF);
	spisend(0xFF);
	
	// release chip select
	SD_PORT |= _BV(SD_CS_PIN);
	
	// return success
	return 1;
}

unsigned char SD_write(unsigned long sector, unsigned char *buffer)
{
	unsigned int i = 0;

	// issue command
	CHECK(SD_command(SD_WRITE_BLOCK, sector << 9) != 0);

	// assert chip select
	SD_PORT &= ~_BV(SD_CS_PIN);
	// send dummy
	spisend(0xFF);
	// send data start token
	spisend(SD_STARTBLOCK_WRITE);
	
	// write data
	for(i = 0; i < SD_BLOCKSIZE; i++)
		spisend(buffer[i]);
	
	// write 16-bit CRC (dummy values)
	spisend(0xFF);
	spisend(0xFF);
	// read data response token
	SD_response = spisend(0xFF);
	CHECK((SD_response & SD_DR_MASK) == SD_DR_ACCEPT);
	
	// wait until card not busy
	while(!spisend(0xFF));
	
	SD_PORT |= _BV(SD_CS_PIN);
	
	// return success
	return 1;
}

#ifdef SD_SPI_CONNECTION
inline unsigned char spisend(unsigned char data)
{
	SPDR = data;
	while(!(SPSR & _BV(SPIF)));
	return SPDR;
}
#else
unsigned char spisend(unsigned char data)
{
	unsigned char i, rcvd = 0; 
	for(i = 0; i < 8; i++)
	{
		if((data >> (7-i)) & 0x01)	
			SD_PORT |= _BV(SD_MOSI_PIN);
		else SD_PORT &= ~_BV(SD_MOSI_PIN);
		
		rcvd |= (((SD_PIN & _BV(SD_MISO_PIN)) >> SD_MISO_PIN) << (7-i));
		
		SD_PORT |= _BV(SD_SCK_PIN);
		SD_PORT &= ~_BV(SD_SCK_PIN);
	}
	return rcvd;
}
#endif
